package java.util.concurrent;

import java.util.NoSuchElementException;
import java.util.Iterator;
import java.util.Collection;
import java.util.Arrays;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;
import java.util.AbstractQueue;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

public class ScheduledThreadPoolExecutor extends ThreadPoolExecutor implements ScheduledExecutorService
{
    private volatile boolean continueExistingPeriodicTasksAfterShutdown;
    private volatile boolean executeExistingDelayedTasksAfterShutdown;
    private volatile boolean removeOnCancel;
    private static final AtomicLong sequencer;
    
    final long now() {
        return System.nanoTime();
    }
    
    boolean canRunInCurrentRunState(boolean b) {
        return this.isRunningOrShutdown(b ? this.continueExistingPeriodicTasksAfterShutdown : this.executeExistingDelayedTasksAfterShutdown);
    }
    
    private void delayedExecute(RunnableScheduledFuture<?> runnableScheduledFuture) {
        if (this.isShutdown()) {
            this.reject(runnableScheduledFuture);
        }
        else {
            super.getQueue().add(runnableScheduledFuture);
            if (this.isShutdown() && !this.canRunInCurrentRunState(runnableScheduledFuture.isPeriodic()) && this.remove(runnableScheduledFuture)) {
                runnableScheduledFuture.cancel(false);
            }
            else {
                this.ensurePrestart();
            }
        }
    }
    
    void reExecutePeriodic(RunnableScheduledFuture<?> runnableScheduledFuture) {
        if (this.canRunInCurrentRunState(true)) {
            super.getQueue().add(runnableScheduledFuture);
            if (!this.canRunInCurrentRunState(true) && this.remove(runnableScheduledFuture)) {
                runnableScheduledFuture.cancel(false);
            }
            else {
                this.ensurePrestart();
            }
        }
    }
    
    @Override
    void onShutdown() {
        BlockingQueue<Runnable> queue = super.getQueue();
        int executeExistingDelayedTasksAfterShutdownPolicy = this.getExecuteExistingDelayedTasksAfterShutdownPolicy();
        int continueExistingPeriodicTasksAfterShutdownPolicy = this.getContinueExistingPeriodicTasksAfterShutdownPolicy();
        if (executeExistingDelayedTasksAfterShutdownPolicy == 0 && continueExistingPeriodicTasksAfterShutdownPolicy == 0) {
            for (Object o2 : queue.toArray()) {
                if ((o2 instanceof RunnableScheduledFuture)) {
                    ((RunnableScheduledFuture)o2).cancel(false);
                }
            }
            queue.clear();
        }
        else {
            for (Object o : queue.toArray()) {
                if ((o instanceof RunnableScheduledFuture)) {
                    RunnableScheduledFuture<Object> runnableScheduledFuture = (RunnableScheduledFuture)o;
                    Label_0164: {
                        if ((!runnableScheduledFuture.isPeriodic()) ? (executeExistingDelayedTasksAfterShutdownPolicy == 0) : (continueExistingPeriodicTasksAfterShutdownPolicy == 0)) {
                            if (queue.remove(runnableScheduledFuture)) {
                                runnableScheduledFuture.cancel(false);
                                continue;
                            }
                            continue;
                        }
                        if (runnableScheduledFuture.isCancelled()) {
                            continue;
                        }
                    }
                }
            }
        }
        this.tryTerminate();
    }
    
    protected <V> RunnableScheduledFuture<V> decorateTask(Runnable runnable, RunnableScheduledFuture<V> runnableScheduledFuture) {
        return runnableScheduledFuture;
    }
    
    protected <V> RunnableScheduledFuture<V> decorateTask(Callable<V> callable, RunnableScheduledFuture<V> runnableScheduledFuture) {
        return runnableScheduledFuture;
    }
    
    public ScheduledThreadPoolExecutor(int n) {
        super(n, Integer.MAX_VALUE, 0L, TimeUnit.NANOSECONDS, new ScheduledThreadPoolExecutor.DelayedWorkQueue());
        this.executeExistingDelayedTasksAfterShutdown = true;
        this.removeOnCancel = false;
    }
    
    public ScheduledThreadPoolExecutor(int n, ThreadFactory threadFactory) {
        super(n, Integer.MAX_VALUE, 0L, TimeUnit.NANOSECONDS, new ScheduledThreadPoolExecutor.DelayedWorkQueue(), threadFactory);
        this.executeExistingDelayedTasksAfterShutdown = true;
        this.removeOnCancel = false;
    }
    
    public ScheduledThreadPoolExecutor(int n, RejectedExecutionHandler rejectedExecutionHandler) {
        super(n, Integer.MAX_VALUE, 0L, TimeUnit.NANOSECONDS, new ScheduledThreadPoolExecutor.DelayedWorkQueue(), rejectedExecutionHandler);
        this.executeExistingDelayedTasksAfterShutdown = true;
        this.removeOnCancel = false;
    }
    
    public ScheduledThreadPoolExecutor(int n, ThreadFactory threadFactory, RejectedExecutionHandler rejectedExecutionHandler) {
        super(n, Integer.MAX_VALUE, 0L, TimeUnit.NANOSECONDS, new ScheduledThreadPoolExecutor.DelayedWorkQueue(), threadFactory, rejectedExecutionHandler);
        this.executeExistingDelayedTasksAfterShutdown = true;
        this.removeOnCancel = false;
    }
    
    private long triggerTime(long n, TimeUnit timeUnit) {
        return this.triggerTime(timeUnit.toNanos((n < 0L) ? 0L : n));
    }
    
    long triggerTime(long n) {
        return this.now() + ((n < 4611686018427387903L) ? n : this.overflowFree(n));
    }
    
    private long overflowFree(long n) {
        Delayed delayed = (Delayed)super.getQueue().peek();
        if (delayed != null) {
            long delay = delayed.getDelay(TimeUnit.NANOSECONDS);
            if (delay < 0L && n - delay < 0L) {
                n = Long.MAX_VALUE + delay;
            }
        }
        return n;
    }
    
    @Override
    public ScheduledFuture<?> schedule(Runnable runnable, long n, TimeUnit timeUnit) {
        if (runnable == null || timeUnit == null) {
            throw new NullPointerException();
        }
        RunnableScheduledFuture<Object> decorateTask = this.decorateTask(runnable, new ScheduledThreadPoolExecutor.ScheduledFutureTask(runnable, null, this.triggerTime(n, timeUnit)));
        this.delayedExecute(decorateTask);
        return decorateTask;
    }
    
    @Override
    public <V> ScheduledFuture<V> schedule(Callable<V> callable, long n, TimeUnit timeUnit) {
        if (callable == null || timeUnit == null) {
            throw new NullPointerException();
        }
        RunnableScheduledFuture<V> decorateTask = this.decorateTask(callable, new ScheduledThreadPoolExecutor.ScheduledFutureTask(callable, this.triggerTime(n, timeUnit)));
        this.delayedExecute(decorateTask);
        return decorateTask;
    }
    
    @Override
    public ScheduledFuture<?> scheduleAtFixedRate(Runnable runnable, long n, long n2, TimeUnit timeUnit) {
        if (runnable == null || timeUnit == null) {
            throw new NullPointerException();
        }
        if (n2 <= 0L) {
            throw new IllegalArgumentException();
        }
        ScheduledThreadPoolExecutor.ScheduledFutureTask scheduledFutureTask = new ScheduledThreadPoolExecutor.ScheduledFutureTask(runnable, null, this.triggerTime(n, timeUnit), timeUnit.toNanos(n2));
        RunnableScheduledFuture<V> decorateTask = this.decorateTask(runnable, scheduledFutureTask);
        this.delayedExecute(scheduledFutureTask.outerTask = decorateTask);
        return decorateTask;
    }
    
    @Override
    public ScheduledFuture<?> scheduleWithFixedDelay(Runnable runnable, long n, long n2, TimeUnit timeUnit) {
        if (runnable == null || timeUnit == null) {
            throw new NullPointerException();
        }
        if (n2 <= 0L) {
            throw new IllegalArgumentException();
        }
        ScheduledThreadPoolExecutor.ScheduledFutureTask scheduledFutureTask = new ScheduledThreadPoolExecutor.ScheduledFutureTask(runnable, null, this.triggerTime(n, timeUnit), timeUnit.toNanos(-n2));
        RunnableScheduledFuture<V> decorateTask = this.decorateTask(runnable, scheduledFutureTask);
        this.delayedExecute(scheduledFutureTask.outerTask = decorateTask);
        return decorateTask;
    }
    
    @Override
    public void execute(Runnable runnable) {
        this.schedule(runnable, 0L, TimeUnit.NANOSECONDS);
    }
    
    @Override
    public Future<?> submit(Runnable runnable) {
        return this.schedule(runnable, 0L, TimeUnit.NANOSECONDS);
    }
    
    @Override
    public <T> Future<T> submit(Runnable runnable, T t) {
        return this.schedule(Executors.callable(runnable, t), 0L, TimeUnit.NANOSECONDS);
    }
    
    @Override
    public <T> Future<T> submit(Callable<T> callable) {
        return this.schedule(callable, 0L, TimeUnit.NANOSECONDS);
    }
    
    public void setContinueExistingPeriodicTasksAfterShutdownPolicy(boolean continueExistingPeriodicTasksAfterShutdown) {
        this.continueExistingPeriodicTasksAfterShutdown = continueExistingPeriodicTasksAfterShutdown;
        if (!continueExistingPeriodicTasksAfterShutdown && this.isShutdown()) {
            this.onShutdown();
        }
    }
    
    public boolean getContinueExistingPeriodicTasksAfterShutdownPolicy() {
        return this.continueExistingPeriodicTasksAfterShutdown;
    }
    
    public void setExecuteExistingDelayedTasksAfterShutdownPolicy(boolean executeExistingDelayedTasksAfterShutdown) {
        this.executeExistingDelayedTasksAfterShutdown = executeExistingDelayedTasksAfterShutdown;
        if (!executeExistingDelayedTasksAfterShutdown && this.isShutdown()) {
            this.onShutdown();
        }
    }
    
    public boolean getExecuteExistingDelayedTasksAfterShutdownPolicy() {
        return this.executeExistingDelayedTasksAfterShutdown;
    }
    
    public void setRemoveOnCancelPolicy(boolean removeOnCancel) {
        this.removeOnCancel = removeOnCancel;
    }
    
    public boolean getRemoveOnCancelPolicy() {
        return this.removeOnCancel;
    }
    
    @Override
    public void shutdown() {
        super.shutdown();
    }
    
    @Override
    public List<Runnable> shutdownNow() {
        return super.shutdownNow();
    }
    
    @Override
    public BlockingQueue<Runnable> getQueue() {
        return super.getQueue();
    }
    
    static {
        sequencer = new AtomicLong();
    }
    
    private class ScheduledFutureTask<V> extends FutureTask<V> implements RunnableScheduledFuture<V>
    {
        private final long sequenceNumber;
        private long time;
        private final long period;
        RunnableScheduledFuture<V> outerTask;
        int heapIndex;
        
        ScheduledFutureTask(Runnable runnable, V v, long time) {
            super(runnable, v);
            this.outerTask = this;
            this.time = time;
            this.period = 0L;
            this.sequenceNumber = ScheduledThreadPoolExecutor.sequencer.getAndIncrement();
        }
        
        ScheduledFutureTask(Runnable runnable, V v, long time, long period) {
            super(runnable, v);
            this.outerTask = this;
            this.time = time;
            this.period = period;
            this.sequenceNumber = ScheduledThreadPoolExecutor.sequencer.getAndIncrement();
        }
        
        ScheduledFutureTask(Callable<V> callable, long time) {
            super(callable);
            this.outerTask = this;
            this.time = time;
            this.period = 0L;
            this.sequenceNumber = ScheduledThreadPoolExecutor.sequencer.getAndIncrement();
        }
        
        @Override
        public long getDelay(TimeUnit timeUnit) {
            return timeUnit.convert(this.time - ScheduledThreadPoolExecutor.this.now(), TimeUnit.NANOSECONDS);
        }
        
        @Override
        public int compareTo(Delayed delayed) {
            if (delayed == this) {
                return 0;
            }
            if ((delayed instanceof ScheduledThreadPoolExecutor.ScheduledFutureTask)) {
                ScheduledThreadPoolExecutor.ScheduledFutureTask<Object> scheduledFutureTask = (ScheduledThreadPoolExecutor.ScheduledFutureTask)delayed;
                long n2 = this.time - scheduledFutureTask.time;
                if (n2 < 0L) {
                    return -1;
                }
                if (n2 > 0L) {
                    return 1;
                }
                if (this.sequenceNumber < scheduledFutureTask.sequenceNumber) {
                    return -1;
                }
                return 1;
            }
            else {
                long n = this.getDelay(TimeUnit.NANOSECONDS) - delayed.getDelay(TimeUnit.NANOSECONDS);
                return (n < 0L) ? -1 : (n > 0L);
            }
        }
        
        @Override
        public boolean isPeriodic() {
            return this.period != 0L;
        }
        
        private void setNextRunTime() {
            long period = this.period;
            if (period > 0L) {
                this.time += period;
            }
            else {
                this.time = ScheduledThreadPoolExecutor.this.triggerTime(-period);
            }
        }
        
        @Override
        public boolean cancel(boolean b) {
            int cancel = super.cancel(b);
            if (cancel && ScheduledThreadPoolExecutor.this.removeOnCancel && this.heapIndex >= 0) {
                ScheduledThreadPoolExecutor.this.remove(this);
            }
            return cancel;
        }
        
        @Override
        public void run() {
            int periodic = this.isPeriodic();
            if (!ScheduledThreadPoolExecutor.this.canRunInCurrentRunState((boolean)periodic)) {
                this.cancel(false);
            }
            else if (periodic == 0) {
                access$201(this);
            }
            else if (access$301(this)) {
                this.setNextRunTime();
                ScheduledThreadPoolExecutor.this.reExecutePeriodic(this.outerTask);
            }
        }
        
        static void access$201(ScheduledThreadPoolExecutor.ScheduledFutureTask scheduledFutureTask) {
            scheduledFutureTask.run();
        }
        
        static boolean access$301(ScheduledThreadPoolExecutor.ScheduledFutureTask scheduledFutureTask) {
            return scheduledFutureTask.runAndReset();
        }
    }
    
    static class DelayedWorkQueue extends AbstractQueue<Runnable> implements BlockingQueue<Runnable>
    {
        private static final int INITIAL_CAPACITY = 16;
        private RunnableScheduledFuture<?>[] queue;
        private final ReentrantLock lock;
        private int size;
        private Thread leader;
        private final Condition available;
        
        DelayedWorkQueue() {
            this.queue = new RunnableScheduledFuture[16];
            this.lock = new ReentrantLock();
            this.size = 0;
            this.leader = null;
            this.available = this.lock.newCondition();
        }
        
        private void setIndex(RunnableScheduledFuture<?> runnableScheduledFuture, int heapIndex) {
            if ((runnableScheduledFuture instanceof ScheduledThreadPoolExecutor.ScheduledFutureTask)) {
                ((ScheduledThreadPoolExecutor.ScheduledFutureTask)runnableScheduledFuture).heapIndex = heapIndex;
            }
        }
        
        private void siftUp(int i, RunnableScheduledFuture<?> runnableScheduledFuture) {
            while (i > 0) {
                int n = i - 1 >>> 1;
                RunnableScheduledFuture<Object> runnableScheduledFuture2 = this.queue[n];
                if (runnableScheduledFuture.compareTo(runnableScheduledFuture2) >= 0) {
                    break;
                }
                this.setIndex(this.queue[i] = runnableScheduledFuture2, i);
                i = n;
            }
            this.setIndex(this.queue[i] = runnableScheduledFuture, i);
        }
        
        private void siftDown(int i, RunnableScheduledFuture<?> runnableScheduledFuture) {
            while (i < this.size >>> 1) {
                int n = (i << 1) + 1;
                RunnableScheduledFuture<Object> runnableScheduledFuture2 = this.queue[n];
                int n2 = n + 1;
                if (n2 < this.size && runnableScheduledFuture2.compareTo(this.queue[n2]) > 0) {
                    runnableScheduledFuture2 = this.queue[n = n2];
                }
                if (runnableScheduledFuture.compareTo(runnableScheduledFuture2) <= 0) {
                    break;
                }
                this.setIndex(this.queue[i] = runnableScheduledFuture2, i);
                i = n;
            }
            this.setIndex(this.queue[i] = runnableScheduledFuture, i);
        }
        
        private void grow() {
            int length = this.queue.length;
            int newLength = length + (length >> 1);
            if (newLength < 0) {
                newLength = Integer.MAX_VALUE;
            }
            this.queue = (RunnableScheduledFuture<?>[])Arrays.copyOf(this.queue, newLength);
        }
        
        private int indexOf(Object o) {
            if (o != null) {
                if ((o instanceof ScheduledThreadPoolExecutor.ScheduledFutureTask)) {
                    int heapIndex = ((ScheduledThreadPoolExecutor.ScheduledFutureTask)o).heapIndex;
                    if (heapIndex >= 0 && heapIndex < this.size && this.queue[heapIndex] == o) {
                        return heapIndex;
                    }
                }
                else {
                    int i = 0;
                    while (i < this.size) {
                        if (o.equals(this.queue[i])) {
                            return i;
                        }
                        ++i;
                    }
                }
            }
            return -1;
        }
        
        @Override
        public boolean contains(Object o) {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                return this.indexOf(o) != -1;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public boolean remove(Object o) {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                int index = this.indexOf(o);
                if (index < 0) {
                    return false;
                }
                this.setIndex(this.queue[index], -1);
                int size = this.size - 1;
                this.size = size;
                int n = size;
                RunnableScheduledFuture<Object> runnableScheduledFuture = this.queue[n];
                this.queue[n] = null;
                if (n != index) {
                    this.siftDown(index, runnableScheduledFuture);
                    if (this.queue[index] == runnableScheduledFuture) {
                        this.siftUp(index, runnableScheduledFuture);
                    }
                }
                return true;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public int size() {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                return this.size;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public boolean isEmpty() {
            return this.size() == 0;
        }
        
        @Override
        public int remainingCapacity() {
            return Integer.MAX_VALUE;
        }
        
        @Override
        public RunnableScheduledFuture<?> peek() {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                return this.queue[0];
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public boolean offer(Runnable runnable) {
            if (runnable == null) {
                throw new NullPointerException();
            }
            RunnableScheduledFuture<Object> runnableScheduledFuture = (RunnableScheduledFuture)runnable;
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                int size = this.size;
                if (size >= this.queue.length) {
                    this.grow();
                }
                this.size = size + 1;
                if (size == 0) {
                    this.setIndex(this.queue[0] = runnableScheduledFuture, 0);
                }
                else {
                    this.siftUp(size, runnableScheduledFuture);
                }
                if (this.queue[0] == runnableScheduledFuture) {
                    this.leader = null;
                    this.available.signal();
                }
            }
            finally {
                lock.unlock();
            }
            return true;
        }
        
        @Override
        public void put(Runnable runnable) {
            this.offer(runnable);
        }
        
        @Override
        public boolean add(Runnable runnable) {
            return this.offer(runnable);
        }
        
        @Override
        public boolean offer(Runnable runnable, long n, TimeUnit timeUnit) {
            return this.offer(runnable);
        }
        
        private RunnableScheduledFuture<?> finishPoll(RunnableScheduledFuture<?> runnableScheduledFuture) {
            int size = this.size - 1;
            this.size = size;
            int n = size;
            RunnableScheduledFuture<Object> runnableScheduledFuture2 = this.queue[n];
            this.queue[n] = null;
            if (n != 0) {
                this.siftDown(0, runnableScheduledFuture2);
            }
            this.setIndex(runnableScheduledFuture, -1);
            return runnableScheduledFuture;
        }
        
        @Override
        public RunnableScheduledFuture<?> poll() {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                RunnableScheduledFuture<Object> runnableScheduledFuture = this.queue[0];
                if (runnableScheduledFuture == null || runnableScheduledFuture.getDelay(TimeUnit.NANOSECONDS) > 0L) {
                    return null;
                }
                return this.finishPoll(runnableScheduledFuture);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public RunnableScheduledFuture<?> take() throws InterruptedException {
            ReentrantLock lock = this.lock;
            lock.lockInterruptibly();
            while (true) {
                try {
                    RunnableScheduledFuture<Object> runnableScheduledFuture = this.queue[0];
                    Block_3: {
                        if (runnableScheduledFuture == null) {
                            this.available.await();
                        }
                        else {
                            long delay = runnableScheduledFuture.getDelay(TimeUnit.NANOSECONDS);
                            if (delay <= 0L) {
                                break Block_3;
                            }
                            if (this.leader != null) {
                                this.available.await();
                            }
                            else {
                                Thread currentThread = Thread.currentThread();
                                this.leader = currentThread;
                                try {
                                    this.available.awaitNanos(delay);
                                }
                                finally {
                                    if (this.leader == currentThread) {
                                        this.leader = null;
                                    }
                                }
                            }
                        }
                        continue;
                    }
                    return this.finishPoll(runnableScheduledFuture);
                }
                finally {
                    if (this.leader == null && this.queue[0] != null) {
                        this.available.signal();
                    }
                    lock.unlock();
                }
            }
        }
        
        @Override
        public RunnableScheduledFuture<?> poll(long n, TimeUnit timeUnit) throws InterruptedException {
            long n2 = timeUnit.toNanos(n);
            ReentrantLock lock = this.lock;
            lock.lockInterruptibly();
            while (true) {
                try {
                    RunnableScheduledFuture<Object> runnableScheduledFuture = this.queue[0];
                    if (runnableScheduledFuture == null) {
                        if (n2 <= 0L) {
                            return null;
                        }
                        n2 = this.available.awaitNanos(n2);
                    }
                    else {
                        long delay = runnableScheduledFuture.getDelay(TimeUnit.NANOSECONDS);
                        if (delay <= 0L) {
                            return this.finishPoll(runnableScheduledFuture);
                        }
                        if (n2 <= 0L) {
                            return null;
                        }
                        if (n2 < delay || this.leader != null) {
                            n2 = this.available.awaitNanos(n2);
                        }
                        else {
                            Thread currentThread = Thread.currentThread();
                            this.leader = currentThread;
                            try {
                                n2 -= delay - this.available.awaitNanos(delay);
                            }
                            finally {
                                if (this.leader == currentThread) {
                                    this.leader = null;
                                }
                            }
                        }
                    }
                    continue;
                }
                finally {
                    if (this.leader == null && this.queue[0] != null) {
                        this.available.signal();
                    }
                    lock.unlock();
                }
            }
        }
        
        @Override
        public void clear() {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                for (int i = 0; i < this.size; ++i) {
                    RunnableScheduledFuture<Object> runnableScheduledFuture = this.queue[i];
                    if (runnableScheduledFuture != null) {
                        this.queue[i] = null;
                        this.setIndex(runnableScheduledFuture, -1);
                    }
                }
                this.size = 0;
            }
            finally {
                lock.unlock();
            }
        }
        
        private RunnableScheduledFuture<?> peekExpired() {
            RunnableScheduledFuture<Object> runnableScheduledFuture = this.queue[0];
            return (runnableScheduledFuture == null || runnableScheduledFuture.getDelay(TimeUnit.NANOSECONDS) > 0L) ? null : runnableScheduledFuture;
        }
        
        @Override
        public int drainTo(Collection<? super Runnable> collection) {
            if (collection == null) {
                throw new NullPointerException();
            }
            if (collection == this) {
                throw new IllegalArgumentException();
            }
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                int n;
                RunnableScheduledFuture<Object> peekExpired;
                for (n = 0; (peekExpired = this.peekExpired()) != null; ++n) {
                    collection.add(peekExpired);
                    this.finishPoll(peekExpired);
                }
                return n;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public int drainTo(Collection<? super Runnable> collection, int n) {
            if (collection == null) {
                throw new NullPointerException();
            }
            if (collection == this) {
                throw new IllegalArgumentException();
            }
            if (n <= 0) {
                return 0;
            }
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                int n2;
                RunnableScheduledFuture<Object> peekExpired;
                for (n2 = 0; n2 < n && (peekExpired = this.peekExpired()) != null; ++n2) {
                    collection.add(peekExpired);
                    this.finishPoll(peekExpired);
                }
                return n2;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public Object[] toArray() {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                return Arrays.copyOf(this.queue, this.size, Object[].class);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public <T> T[] toArray(T[] array) {
            ReentrantLock lock = this.lock;
            lock.lock();
            try {
                if (array.length < this.size) {
                    return (T[])Arrays.copyOf(this.queue, this.size, array.getClass());
                }
                System.arraycopy(this.queue, 0, array, 0, this.size);
                if (array.length > this.size) {
                    array[this.size] = null;
                }
                return array;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public Iterator<Runnable> iterator() {
            return new ScheduledThreadPoolExecutor.DelayedWorkQueue.Itr((RunnableScheduledFuture<?>[])Arrays.copyOf(this.queue, this.size));
        }
        
        private class Itr implements Iterator<Runnable>
        {
            final RunnableScheduledFuture<?>[] array;
            int cursor;
            int lastRet;
            
            Itr(RunnableScheduledFuture<?>[] array) {
                this.cursor = 0;
                this.lastRet = -1;
                this.array = array;
            }
            
            @Override
            public boolean hasNext() {
                return this.cursor < this.array.length;
            }
            
            @Override
            public Runnable next() {
                if (this.cursor >= this.array.length) {
                    throw new NoSuchElementException();
                }
                this.lastRet = this.cursor;
                return this.array[this.cursor++];
            }
            
            @Override
            public void remove() {
                if (this.lastRet < 0) {
                    throw new IllegalStateException();
                }
                ScheduledThreadPoolExecutor.DelayedWorkQueue.this.remove(this.array[this.lastRet]);
                this.lastRet = -1;
                return;
            }
        }
    }
}
